@@ -106,16 +106,16 @@ _jumpstart_other_acmds (mongoc_topology_scanner_node_t *node,
 static void
 _add_hello (mongoc_topology_scanner_t *ts)
 {
-   mongoc_server_api_t *api = ts->api;
-
    BSON_APPEND_INT32 (&ts->hello_cmd, "hello", 1);
    BSON_APPEND_BOOL (&ts->hello_cmd, "helloOk", true);
 
    BSON_APPEND_INT32 (&ts->legacy_hello_cmd, HANDSHAKE_CMD_LEGACY_HELLO, 1);
    BSON_APPEND_BOOL (&ts->legacy_hello_cmd, "helloOk", true);
 
-   if (api) {
-      _mongoc_cmd_append_server_api (&ts->hello_cmd, api);
+   /* Append appropriate server API metadata (such as "serverApi") if selected:
+    */
+   if (mongoc_topology_scanner_uses_server_api (ts)) {
+      _mongoc_cmd_append_server_api (&ts->hello_cmd, ts->api);
    }
 }
 
@@ -124,8 +124,8 @@ _init_hello (mongoc_topology_scanner_t *ts)
 {
    bson_init (&ts->hello_cmd);
    bson_init (&ts->legacy_hello_cmd);
-   bson_init (&ts->handshake_cmd);
    bson_init (&ts->cluster_time);
+   ts->handshake_cmd = NULL;
 
    _add_hello (ts);
 }
@@ -133,9 +133,16 @@ _init_hello (mongoc_topology_scanner_t *ts)
 static void
 _reset_hello (mongoc_topology_scanner_t *ts)
 {
+   bson_t *prev_cmd;
    bson_reinit (&ts->hello_cmd);
    bson_reinit (&ts->legacy_hello_cmd);
-   bson_reinit (&ts->handshake_cmd);
+
+   bson_mutex_lock (&ts->handshake_cmd_mtx);
+   prev_cmd = ts->handshake_cmd;
+   ts->handshake_cmd = NULL;
+   ts->handshake_state = HANDSHAKE_CMD_UNINITIALIZED;
+   bson_mutex_unlock (&ts->handshake_cmd_mtx);
+   bson_destroy (prev_cmd);
 
    _add_hello (ts);
 }
@@ -244,29 +251,37 @@ _mongoc_topology_scanner_parse_speculative_authentication (
    bson_copy_to (&auth_response, speculative_authenticate);
 }
 
-static bool
-_build_handshake_cmd (mongoc_topology_scanner_t *ts)
+static bson_t *
+_build_handshake_cmd (const bson_t *basis_cmd,
+                      const char *appname,
+                      const mongoc_uri_t *uri,
+                      bool is_loadbalanced)
 {
-   bson_t *doc = &ts->handshake_cmd;
+   bson_t *doc = bson_copy (basis_cmd);
    bson_t subdoc;
    bson_iter_t iter;
    const char *key;
    int keylen;
-   bool res;
    const bson_t *compressors;
    int count = 0;
    char buf[16];
+   bool subdoc_okay;
 
-   bson_destroy (doc);
-   bson_copy_to (ts->api ? &ts->hello_cmd : &ts->legacy_hello_cmd, doc);
+   BSON_ASSERT (doc);
 
    BSON_APPEND_DOCUMENT_BEGIN (doc, HANDSHAKE_FIELD, &subdoc);
-   res = _mongoc_handshake_build_doc_with_application (&subdoc, ts->appname);
+   subdoc_okay =
+      _mongoc_handshake_build_doc_with_application (&subdoc, appname);
    bson_append_document_end (doc, &subdoc);
 
+   if (!subdoc_okay) {
+      bson_destroy (doc);
+      return NULL;
+   }
+
    BSON_APPEND_ARRAY_BEGIN (doc, "compression", &subdoc);
-   if (ts->uri) {
-      compressors = mongoc_uri_get_compressors (ts->uri);
+   if (uri) {
+      compressors = mongoc_uri_get_compressors (uri);
 
       if (bson_iter_init (&iter, compressors)) {
          while (bson_iter_next (&iter)) {
@@ -278,39 +293,87 @@ _build_handshake_cmd (mongoc_topology_scanner_t *ts)
    }
    bson_append_array_end (doc, &subdoc);
 
+   if (is_loadbalanced) {
+      BSON_APPEND_BOOL (doc, "loadBalanced", true);
+   }
+
    /* Return whether the handshake doc fit the size limit */
-   return res;
+   return doc;
 }
 
 const bson_t *
 _mongoc_topology_scanner_get_monitoring_cmd (mongoc_topology_scanner_t *ts,
                                              bool hello_ok)
 {
-   return hello_ok || ts->api ? &ts->hello_cmd : &ts->legacy_hello_cmd;
+   return hello_ok || mongoc_topology_scanner_uses_server_api (ts)
+             ? &ts->hello_cmd
+             : &ts->legacy_hello_cmd;
 }
 
-/* Caller must lock topology->mutex to protect handshake_cmd. This
- * is called at the start of the scan in _mongoc_topology_run_background, when a
- * node is added in _mongoc_topology_reconcile_add_nodes, or when running a
- * hello directly on a node in _mongoc_stream_run_hello. */
-const bson_t *
-_mongoc_topology_scanner_get_handshake_cmd (mongoc_topology_scanner_t *ts)
+void
+_mongoc_topology_scanner_dup_handshake_cmd (mongoc_topology_scanner_t *ts,
+                                            bson_t *copy_into)
 {
+   bson_t *new_cmd;
+   const char *appname;
+   BSON_ASSERT_PARAM (ts);
+   BSON_ASSERT_PARAM (copy_into);
+
+   /* appname will only be changed from NULL, so a non-null pointer will never
+    * be invalidated after this fetch. */
+   appname =
+      bson_atomic_ptr_fetch ((void *) &ts->appname, bson_memory_order_relaxed);
+
+   bson_mutex_lock (&ts->handshake_cmd_mtx);
    /* If this is the first time using the node or if it's the first time
     * using it after a failure, build handshake doc */
-   if (bson_empty (&ts->handshake_cmd)) {
-      ts->handshake_ok_to_send = _build_handshake_cmd (ts);
-      if (!ts->handshake_ok_to_send) {
-         MONGOC_WARNING ("Handshake doc too big, not including in hello");
-      }
+   if (ts->handshake_state != HANDSHAKE_CMD_UNINITIALIZED) {
+      /* We're good to just return the handshake now */
+      goto after_init;
    }
 
+   /* There is not yet a handshake command associated with this scanner.
+    * Initialize one and set it now. */
+   /* Note: Don't hold the mutex while we build our command */
+   /* Construct a new handshake command to be sent */
+   BSON_ASSERT (ts->handshake_cmd == NULL);
+   bson_mutex_unlock (&ts->handshake_cmd_mtx);
+   new_cmd = _build_handshake_cmd (mongoc_topology_scanner_uses_server_api (ts)
+                                      ? &ts->hello_cmd
+                                      : &ts->legacy_hello_cmd,
+                                   appname,
+                                   ts->uri,
+                                   ts->loadbalanced);
+   bson_mutex_lock (&ts->handshake_cmd_mtx);
+   if (ts->handshake_state != HANDSHAKE_CMD_UNINITIALIZED) {
+      /* Someone else updated the handshake_cmd while we were building ours.
+       * Defer to their copy and just destroy the one we created. */
+      bson_destroy (new_cmd);
+      goto after_init;
+   }
+   BSON_ASSERT (ts->handshake_cmd == NULL);
+   /* We're still the one updating the command */
+   ts->handshake_cmd = new_cmd;
+   /* The "_build" may have failed. */
+   /* Even if new_cmd is NULL, this is still what we want */
+   ts->handshake_state =
+      new_cmd == NULL ? HANDSHAKE_CMD_TOO_BIG : HANDSHAKE_CMD_OKAY;
+   if (ts->handshake_state == HANDSHAKE_CMD_TOO_BIG) {
+      MONGOC_WARNING ("Handshake doc too big, not including in hello");
+   }
+
+after_init:
    /* If the doc turned out to be too big */
-   if (!ts->handshake_ok_to_send) {
-      return ts->api ? &ts->hello_cmd : &ts->legacy_hello_cmd;
+   if (ts->handshake_state == HANDSHAKE_CMD_TOO_BIG) {
+      bson_t *ret = mongoc_topology_scanner_uses_server_api (ts)
+                       ? &ts->hello_cmd
+                       : &ts->legacy_hello_cmd;
+      bson_copy_to (ret, copy_into);
+   } else {
+      BSON_ASSERT (ts->handshake_cmd != NULL);
+      bson_copy_to (ts->handshake_cmd, copy_into);
    }
-
-   return &ts->handshake_cmd;
+   bson_mutex_unlock (&ts->handshake_cmd_mtx);
 }
 
 static void
@@ -318,17 +381,26 @@ _begin_hello_cmd (mongoc_topology_scanner_node_t *node,
                   mongoc_stream_t *stream,
                   bool is_setup_done,
                   struct addrinfo *dns_result,
-                  int64_t initiate_delay_ms)
+                  int64_t initiate_delay_ms,
+                  bool use_handshake)
 {
    mongoc_topology_scanner_t *ts = node->ts;
+   mongoc_opcode_t cmd_opcode_type = MONGOC_OPCODE_QUERY;
    bson_t cmd;
 
-   if (node->last_used != -1 && node->last_failed == -1) {
+   /* If we're asked to use a specific API version, we should send our
+   hello handshake via op_msg rather than the legacy op_query: */
+   if (mongoc_topology_scanner_uses_server_api (ts)) {
+      cmd_opcode_type = MONGOC_OPCODE_MSG;
+   }
+
+   if (node->last_used != -1 && node->last_failed == -1 && !use_handshake) {
       /* The node's been used before and not failed recently */
-      bson_copy_to (_mongoc_topology_scanner_get_monitoring_cmd (ts, node->hello_ok),
-                    &cmd);
+      bson_copy_to (
+         _mongoc_topology_scanner_get_monitoring_cmd (ts, node->hello_ok),
+         &cmd);
    } else {
-      bson_copy_to (_mongoc_topology_scanner_get_handshake_cmd (ts), &cmd);
+      _mongoc_topology_scanner_dup_handshake_cmd (ts, &cmd);
    }
 
    if (node->ts->negotiate_sasl_supported_mechs &&
@@ -344,6 +416,10 @@ _begin_hello_cmd (mongoc_topology_scanner_node_t *node,
       ssl_opts = ts->ssl_opts;
 #endif
 
+      // _mongoc_topology_scanner_add_speculative_authentication is called with
+      // NULL for the scram_cache argument. The scram cache is not used for
+      // speculative authentication in the topology scanner. This is planned to
+      // be improved in CDRIVER-3642.
       _mongoc_topology_scanner_add_speculative_authentication (
          &cmd, ts->uri, ssl_opts, NULL, &node->scram);
    }
@@ -365,6 +441,7 @@ _begin_hello_cmd (mongoc_topology_scanner_node_t *node,
                          node->host.host,
                          "admin",
                          &cmd,
+                         cmd_opcode_type,
                          &_async_handler,
                          node,
                          ts->connect_timeout_msec);
@@ -382,7 +459,7 @@ mongoc_topology_scanner_new (
    int64_t connect_timeout_msec)
 {
    mongoc_topology_scanner_t *ts =
-      (mongoc_topology_scanner_t *) bson_malloc0 (sizeof (*ts));
+      BSON_ALIGNED_ALLOC0 (mongoc_topology_scanner_t);
 
    ts->async = mongoc_async_new ();
 
@@ -392,10 +469,11 @@ mongoc_topology_scanner_new (
    ts->uri = uri;
    ts->appname = NULL;
    ts->api = NULL;
-   ts->handshake_ok_to_send = false;
+   ts->handshake_state = HANDSHAKE_CMD_UNINITIALIZED;
    ts->connect_timeout_msec = connect_timeout_msec;
    /* may be overridden for testing. */
    ts->dns_cache_timeout_ms = DNS_CACHE_TIMEOUT_MS;
+   bson_mutex_init (&ts->handshake_cmd_mtx);
 
    _init_hello (ts);
 
@@ -435,9 +513,10 @@ mongoc_topology_scanner_destroy (mongoc_topology_scanner_t *ts)
    mongoc_async_destroy (ts->async);
    bson_destroy (&ts->hello_cmd);
    bson_destroy (&ts->legacy_hello_cmd);
-   bson_destroy (&ts->handshake_cmd);
+   bson_destroy (ts->handshake_cmd);
    bson_destroy (&ts->cluster_time);
    mongoc_server_api_destroy (ts->api);
+   bson_mutex_destroy (&ts->handshake_cmd_mtx);
 
    /* This field can be set by a mongoc_client */
    bson_free ((char *) ts->appname);
@@ -461,7 +540,7 @@ mongoc_topology_scanner_add (mongoc_topology_scanner_t *ts,
 {
    mongoc_topology_scanner_node_t *node;
 
-   node = (mongoc_topology_scanner_node_t *) bson_malloc0 (sizeof (*node));
+   node = BSON_ALIGNED_ALLOC0 (mongoc_topology_scanner_node_t);
 
    memcpy (&node->host, host, sizeof (*host));
 
@@ -526,11 +605,9 @@ mongoc_topology_scanner_node_disconnect (mongoc_topology_scanner_node_t *node,
       }
 
       node->stream = NULL;
-      memset (
-         &node->sasl_supported_mechs, 0, sizeof (node->sasl_supported_mechs));
-      node->negotiated_sasl_supported_mechs = false;
-      bson_reinit (&node->speculative_auth_response);
    }
+   mongoc_server_description_destroy (node->handshake_sd);
+   node->handshake_sd = NULL;
 }
 
 void
@@ -643,6 +720,17 @@ _async_success (mongoc_async_cmd_t *acmd,
    BSON_ASSERT (!node->stream);
    node->stream = stream;
 
+   if (!node->handshake_sd) {
+      mongoc_server_description_t sd;
+
+      /* Store a server description associated with the handshake. */
+      mongoc_server_description_init (&sd, node->host.host_and_port, node->id);
+      mongoc_server_description_handle_hello (
+         &sd, hello_response, duration_usec / 1000, &acmd->error);
+      node->handshake_sd = mongoc_server_description_new_copy (&sd);
+      mongoc_server_description_cleanup (&sd);
+   }
+
    if (ts->negotiate_sasl_supported_mechs &&
        !node->negotiated_sasl_supported_mechs) {
       _mongoc_handshake_parse_sasl_supported_mechs (
@@ -716,6 +804,9 @@ _async_error_or_timeout (mongoc_async_cmd_t *acmd,
       /* call the topology scanner callback. cannot connect to this node.
        * callback takes rtt_msec, not usec. */
       ts->cb (node->id, NULL, duration_usec / 1000, ts->cb_data, error);
+
+      mongoc_server_description_destroy (node->handshake_sd);
+      node->handshake_sd = NULL;
    } else {
       /* there are still more commands left for this node or it succeeded
        * with another stream. skip the topology scanner callback. */
@@ -871,11 +962,21 @@ mongoc_topology_scanner_node_setup_tcp (mongoc_topology_scanner_node_t *node,
    }
 
    if (node->successful_dns_result) {
-      _begin_hello_cmd (node, NULL, false, node->successful_dns_result, 0);
+      _begin_hello_cmd (node,
+                        NULL /* stream */,
+                        false /* is_setup_done */,
+                        node->successful_dns_result,
+                        0 /* initiate_delay_ms */,
+                        true /* use_handshake */);
    } else {
       LL_FOREACH2 (node->dns_results, iter, ai_next)
       {
-         _begin_hello_cmd (node, NULL, false, iter, delay);
+         _begin_hello_cmd (node,
+                           NULL /* stream */,
+                           false /* is_setup_done */,
+                           iter,
+                           delay,
+                           true /* use_handshake */);
          /* each subsequent DNS result will have an additional 250ms delay. */
          delay += HAPPY_EYEBALLS_DELAY_MS;
       }
@@ -942,7 +1043,8 @@ mongoc_topology_scanner_node_connect_unix (mongoc_topology_scanner_node_t *node,
                         stream,
                         false /* is_setup_done */,
                         NULL /* dns result */,
-                        0 /* delay */);
+                        0 /* delay */,
+                        true /* use_handshake */);
       RETURN (true);
    }
    bson_set_error (error,
@@ -980,19 +1082,40 @@ mongoc_topology_scanner_node_setup (mongoc_topology_scanner_node_t *node,
 
    /* if there is already a working stream, push it back to be re-scanned. */
    if (node->stream) {
-      _begin_hello_cmd (node, node->stream, true /* is_setup_done */, NULL, 0);
+      _begin_hello_cmd (node,
+                        node->stream,
+                        true /* is_setup_done */,
+                        NULL /* dns_result */,
+                        0 /* initiate_delay_ms */,
+                        false /* use_handshake */);
       node->stream = NULL;
       return;
    }
 
    BSON_ASSERT (!node->retired);
 
+   // If a new stream is needed, reset state authentication state.
+   // Authentication state is tied to a stream.
+   {
+      node->has_auth = false;
+      bson_reinit (&node->speculative_auth_response);
+      node->scram.step = 0;
+      memset (
+         &node->sasl_supported_mechs, 0, sizeof (node->sasl_supported_mechs));
+      node->negotiated_sasl_supported_mechs = false;
+   }
+
    if (node->ts->initiator) {
       stream = node->ts->initiator (
          node->ts->uri, &node->host, node->ts->initiator_context, error);
       if (stream) {
          success = true;
-         _begin_hello_cmd (node, stream, false, NULL, 0);
+         _begin_hello_cmd (node,
+                           stream,
+                           false /* is_setup_done */,
+                           NULL /* dns_result */,
+                           0 /* initiate_delay_ms */,
+                           true /* use_handshake */);
       }
    } else {
       if (node->host.family == AF_UNIX) {
@@ -1012,8 +1135,6 @@ mongoc_topology_scanner_node_setup (mongoc_topology_scanner_node_t *node,
       node->ts->setup_err_cb (node->id, node->ts->cb_data, error);
       return;
    }
-
-   node->has_auth = false;
 }
 
 /*
@@ -1084,8 +1205,6 @@ mongoc_topology_scanner_in_cooldown (mongoc_topology_scanner_t *ts,
  *      should be called once before calling mongoc_topology_scanner_work()
  *      to complete the scan.
  *
- *      The topology mutex must be held by the caller.
- *
  *      If "obey_cooldown" is true, this is a single-threaded blocking scan
  *      that must obey the Server Discovery And Monitoring Spec's cooldownMS:
  *
@@ -1215,18 +1334,23 @@ bool
 _mongoc_topology_scanner_set_appname (mongoc_topology_scanner_t *ts,
                                       const char *appname)
 {
+   char *s;
+   const char *prev;
    if (!_mongoc_handshake_appname_is_valid (appname)) {
       MONGOC_ERROR ("Cannot set appname: %s is invalid", appname);
       return false;
    }
 
-   if (ts->appname != NULL) {
-      MONGOC_ERROR ("Cannot set appname more than once");
-      return false;
+   s = bson_strdup (appname);
+   prev = bson_atomic_ptr_compare_exchange_strong (
+      (void *) &ts->appname, NULL, s, bson_memory_order_relaxed);
+   if (prev == NULL) {
+      return true;
    }
 
-   ts->appname = bson_strdup (appname);
-   return true;
+   MONGOC_ERROR ("Cannot set appname more than once");
+   bson_free (s);
+   return false;
 }
 
 /*
@@ -1366,15 +1490,31 @@ _jumpstart_other_acmds (mongoc_topology_scanner_node_t *node,
    }
 }
 
-/* Caller must lock topology->mutex to protect handshake_cmd. */
 void
 _mongoc_topology_scanner_set_server_api (mongoc_topology_scanner_t *ts,
                                          const mongoc_server_api_t *api)
 {
    BSON_ASSERT (ts);
    BSON_ASSERT (api);
-
    mongoc_server_api_destroy (ts->api);
    ts->api = mongoc_server_api_copy (api);
    _reset_hello (ts);
 }
+
+/* This must be called before the handshake command is constructed. */
+void
+_mongoc_topology_scanner_set_loadbalanced (mongoc_topology_scanner_t *ts,
+                                           bool val)
+{
+   BSON_UNUSED (val);
+
+   BSON_ASSERT (ts->handshake_cmd == NULL);
+   ts->loadbalanced = true;
+}
+
+bool
+mongoc_topology_scanner_uses_server_api (
+   const mongoc_topology_scanner_t *topology_scanner)
+{
+   return NULL != topology_scanner->api;
+}
